"""Test the localization model.
"""

from sklearn.pipeline import make_pipeline
import numpy as np
import dask.array as da
from pymks.fmks.bases.primitive import discretize, redundancy
from pymks.fmks.localization import fit
from pymks.fmks.bases.primitive import PrimitiveTransformer
from pymks.fmks.localization import LocalizationRegressor
from pymks.fmks.data.delta import generate
from pymks.fmks.data.elastic_fe import solve


def _get_x():
    return da.from_array(np.linspace(0, 1, 4).reshape((1, 2, 2)), chunks=(1, 2, 2))


def test():
    """Very simple example.
    """
    assert np.allclose(
        fit(
            _get_x(),
            _get_x().swapaxes(1, 2),
            discretize(n_state=2),
            redundancy_func=redundancy,
        ),
        [[[0.5, 0.5], [-2, 0]], [[-0.5, 0], [-1, 0]]],
    )


def test_setting_kernel():
    """Test resetting the coeffs after coeff resize.
    """

    x_data = generate(n_phases=2, shape=(21, 21)).persist()

    y_data = solve(
        x_data, elastic_modulus=(100, 130), poissons_ratio=(0.3, 0.3), macro_strain=0.01
    )["strain"][..., 0].persist()

    model = make_pipeline(PrimitiveTransformer(n_state=2), LocalizationRegressor())

    shape = (30, 30)
    fcoeff = model.fit(x_data, y_data).steps[1][1].coeff
    assert np.allclose(model.steps[1][1].coeff_resize(shape).coeff.shape[:-1], shape)
    model.steps[1][1].coeff = fcoeff
    assert np.allclose(model.predict(x_data), y_data, atol=1e-4)



"""fMKS - functional matierals knowledge system.

See https://github.com/wd15/fmks

fMKS is a functional version of [PyMKS](https://pymks.org) currently
under development. The purpose of the project is to prototype a
parallel implementation of MKS using functional programming in Python
primarily using the [Toolz](http://toolz.readthedocs.io) library.

"""

import os
import pytest
from pkg_resources import get_distribution, DistributionNotFound


def test():  # pragma: no cover
    r"""
    Run all the doctests available.
    """

    path = os.path.split(__file__)[0]
    pytest.main(
        args=[path, "--doctest-modules", "--ignore=setup.py ", "-r s", "--cov=fmks"]
    )


def get_version() -> str:
    """Get the version of the code from egg_info.

    Returns:
      the package version number
    """

    try:
        # pylint: disable=no-member
        version = get_distribution(__name__.split(".")[0]).version
    except DistributionNotFound:  # pragma: no cover
        version = "unknown, try running `python setup.py egg_info`"

    return version


__version__ = get_version()

__all__ = ["__version__", "test"]



"""Plotting functions for PyMKS notebooks
"""
import matplotlib.pyplot as plt
import numpy
from .func import curry, fmap


@curry
def _plot_ax(axis, arrs, titles, cmap):
    axis.set_xticks(())
    axis.set_yticks(())
    axis.set_title(titles[axis.colNum])
    return axis.imshow(
        arrs[axis.colNum],
        interpolation="none",
        vmin=numpy.min(numpy.vstack(arrs)),
        vmax=numpy.max(numpy.vstack(arrs)),
        cmap=cmap,
    )


def _colorbar(fig, axis, image):
    """Generate a colorbar
    """
    axis.yaxis.set_offset_position("right")
    fig.colorbar(image, cax=axis)


def plot_microstructures(*arrs, titles=(), cmap=None, colorbar=True):
    """Plot a set of microstructures

    Args:
      *arrs: any number of 2D arrays to plot
      titles: a sequence of titles with len(*arrs)
      cmap: any matplotlib colormap

    """
    fig, axs = plt.subplots(
        1, len(arrs), figsize=(4 * len(arrs), 4), constrained_layout=True
    )
    if len(arrs) == 1:
        axs = (axs,)
    if isinstance(titles, str):
        titles = (titles,)
    if len(titles) < len(arrs):
        titles = titles + ("",) * (len(arrs) - len(titles))
    plots = list(fmap(_plot_ax(arrs=arrs, titles=titles, cmap=cmap), axs))
    if colorbar:
        _colorbar(
            fig, fig.add_axes([1.0, 0.05, 0.05, 0.9]), plots[0],
        )



"""PyMKS - the materials knowledge system in Python

See the documenation for details at https://pymks.org
"""

try:
    import pyfftw  # noqa: F401

    # ensure that pyfftw is always imported before numpy to avoid
    # https://github.com/materialsinnovation/pymks/issues/304
except ImportError:
    pass

import os


from .fmks.data.cahn_hilliard import solve as solve_cahn_hilliard
from .fmks.plot import plot_microstructures
from .fmks.bases.primitive import PrimitiveTransformer
from .fmks.bases.legendre import LegendreTransformer
from .fmks.localization import LocalizationRegressor
from .fmks.localization import ReshapeTransformer
from .fmks.localization import coeff_to_real
from .fmks.data.delta import generate as generate_delta

try:
    import sfepy  # noqa: F401
except ImportError:

    def solve_fe(*_, **__):
        """Dummy funcion when sfepy unavailable
        """
        # pylint: disable=redefined-outer-name, import-outside-toplevel, unused-import
        import sfepy  # noqa: F401, F811


else:
    from .fmks.data.elastic_fe import solve as solve_fe
from .fmks.data.multiphase import generate as generate_multiphase
from .fmks.correlations import FlattenTransformer
from .fmks.correlations import TwoPointCorrelation
from .fmks.data.checkerboard import generate as generate_checkerboard

from .mks_localization_model import MKSLocalizationModel
from .bases.primitive import PrimitiveBasis
from .bases.legendre import LegendreBasis
from .mks_structure_analysis import MKSStructureAnalysis
from .mks_homogenization_model import MKSHomogenizationModel

MKSRegressionModel = MKSLocalizationModel
DiscreteIndicatorBasis = PrimitiveBasis
ContinuousIndicatorBasis = PrimitiveBasis


def test():
    r"""
    Run all the doctests available.
    """
    import pytest  # pylint: disable=import-outside-toplevel

    path = os.path.split(__file__)[0]
    pytest.main(args=[path, "--doctest-modules", "-r s"])


def get_version():
    """Get the version of the code from egg_info.

    Returns:
      the package version number
    """
    # pylint: disable=import-outside-toplevel
    from pkg_resources import get_distribution, DistributionNotFound

    try:
        version = get_distribution(
            __name__.split(".")[0]
        ).version  # pylint: disable=no-member
    except DistributionNotFound:  # pragma: no cover
        version = "unknown, try running `python setup.py egg_info`"

    return version


__version__ = get_version()

__all__ = [
    "__version__",
    "test",
    "solve_cahn_hilliard",
    "plot_microstructures",
    "PrimitiveTransformer",
    "LocalizationRegressor",
    "ReshapeTransformer",
    "coeff_to_real",
    "MKSLocalizationModel",
    "PrimitiveBasis",
    "LegendreBasis",
    "MKSHomogenizationModel",
    "MKSStructureAnalysis",
    "generate_delta",
    "LegendreTransformer",
    "solve_fe",
    "generate_multiphase",
    "FlattenTransformer",
    "TwoPointCorrelation",
    "generate_checkerboard",
]



"""Generate a delta microstructure for training elastic PyMKS models.

"""

from numpy import arange, transpose, array, identity
import dask.array as da

from ..func import npresize, extend, assign, sequence, curry


@curry
def _npgenerate(n_phases, shape):
    return sequence(
        arange,
        npresize((n_phases, n_phases)),
        transpose,
        extend(shape),
        assign(
            arange(n_phases)[None, :],
            ((slice(None), slice(None)) + tuple(array(shape) // 2)),
        ),
        lambda x: x[~identity(n_phases, dtype=bool)],
    )(n_phases)


@curry
def generate(n_phases, shape, chunks=()):
    """Generate a delta microstructure

    Args:
      n_phases: number of phases
      shape: the shape of the microstructure
      chunks: how to chunk the sample axis

    Returns:
      a dask array of delta microstructures

    >>> a = generate(5, (3, 4), chunks=(5,))
    >>> a.shape
    (20, 3, 4)
    >>> a.chunks
    ((5, 5, 5, 5), (3,), (4,))
    """
    return da.from_array(_npgenerate(n_phases, shape), chunks=(chunks or (-1,)) + shape)



"""MKS Correlation Module
For computing auto and cross corelations under assumption
of  periodic or  non-periodic boundary conditions using discreete fourier
transform.


Note that input microstrucure should be 4 dimensional array.
where X=[n_sample,x,y.n_basis]
"""
import numpy as np
from toolz.curried import pipe, curry
from toolz.curried import map as map_, identity
from sklearn.base import TransformerMixin, BaseEstimator
import dask.array as da
from .func import dafftshift, dafftn, daifftn, daconj, flatten
from .func import sequence, make_da, star, dapad


@make_da
def cross_correlation(arr1, arr2):
    """
    Returns the cross-correlation of and input field with another field.

    Args:
      arr1: the input field (n_samples,n_x,n_y)
      arr2: the other input field (n_samples,n_x,n_y)

    Returns:
      an nd-array of same dimension as the input field

    >>> import dask.array as da
    >>> x_data = np.asarray([[[1,1,0],
    ...                       [0,0,1],
    ...                       [1,1,0]]])
    >>> chunks = x_data.shape
    >>> x_data = da.from_array(x_data, chunks=chunks)
    >>> y_data = 1 - x_data
    >>> f_data = cross_correlation(x_data, y_data)
    >>> gg = np.asarray([[[ 2/9,  3/9,  2/9],
    ...                   [ 3/9, 0,  3/9],
    ...                   [ 2/9,  3/9,  2/9]]])
    >>> assert np.allclose(f_data.compute(), gg)
    >>> da.random.seed(42)
    >>> shape = (10, 5, 5)
    >>> chunks = (2, 5, 5)
    >>> x_data = da.random.random(shape, chunks=chunks)
    >>> y_data = 1 - x_data
    >>> f_data = cross_correlation(x_data, y_data)
    >>> assert x_data.chunks == f_data.chunks
    >>> shape = (10, 5, 5)
    >>> # When the two input fields have different chunkings
    >>> x_data = da.random.random(shape, chunks=(2,5,5))
    >>> y_data = da.random.random(shape, chunks=(5,5,5))
    >>> f_data = cross_correlation(x_data, y_data)
    >>> print(f_data.chunks)
    ((2, 2, 1, 1, 2, 2), (5,), (5,))
    """
    faxes = lambda x: tuple(np.arange(x.ndim - 1) + 1)

    return pipe(
        arr1,
        dafftn(axes=faxes(arr1)),
        lambda x: daconj(x) * dafftn(arr2, axes=faxes(arr2)),
        daifftn(axes=faxes(arr1)),
        dafftshift(axes=faxes(arr1)),
        lambda x: x.real / arr1[0].size,
    )


@curry
def auto_correlation(arr):
    """
    Returns auto-corrlation of and input field with itself.

    Args:
      arr: the input field (n_samples,n_x,n_y)

    Returns:
      an nd-array of same dimension as the input field

    >>> import dask.array as da
    >>> x_data = np.asarray([[[1, 1, 0],
    ...                       [0, 0, 1],
    ...                       [1, 1, 0]]])
    >>> chunks = x_data.shape
    >>> x_data = da.from_array(x_data, chunks=chunks)
    >>> f_data = auto_correlation(x_data)
    >>> gg = [[[3/9, 2/9, 3/9],
    ...        [2/9, 5/9, 2/9],
    ...        [3/9, 2/9, 3/9]]]
    >>> assert np.allclose(f_data.compute(), gg)
    >>> shape = (7, 5, 5)
    >>> chunks = (2, 5, 5)
    >>> da.random.seed(42)
    >>> x_data = da.random.random(shape, chunks=chunks)
    >>> f_data = auto_correlation(x_data)
    >>> assert x_data.chunks == f_data.chunks
    >>> print(f_data.chunks)
    ((2, 2, 2, 1), (5,), (5,))
    """
    return cross_correlation(arr, arr)


@curry
def center_slice(x_data, cutoff):
    """Calculate region of interest around the center voxel upto the
    cutoff length

    Args:
      x_data: the data array (n_samples,n_x,n_y), first index is left unchanged
      cutoff: cutoff size

    Returns:
      reduced size array

    >>> a = np.arange(7).reshape(1, 7)
    >>> print(center_slice(a, 2))
    [[1 2 3 4 5]]

    >>> a = np.arange(49).reshape(1, 7, 7)
    >>> print(center_slice(a, 1).shape)
    (1, 3, 3)

    >>> center_slice(np.arange(5), 1)
    Traceback (most recent call last):
    ...
    RuntimeError: Data should be greater than 1D

    """
    if x_data.ndim <= 1:
        raise RuntimeError("Data should be greater than 1D")

    make_slice = sequence(
        lambda x: x_data.shape[1:][x] // 2, lambda x: slice(x - cutoff, x + cutoff + 1)
    )

    return pipe(
        range(len(x_data.shape) - 1),
        map_(make_slice),
        tuple,
        lambda x: (slice(len(x_data)),) + x,
        lambda x: x_data[x],
    )


@curry
def two_point_stats(arr1, arr2, periodic_boundary=True, cutoff=None):
    """Calculate the 2-points stats for two arrays

    Args:
      arr1: array used to calculate cross-correlations (n_samples,n_x,n_y)
      arr2: array used to calculate cross-correlations (n_samples,n_x,n_y)
      periodic_boundary: whether to assume a periodic boundary (default is true)
      cutoff: the subarray of the 2 point stats to keep

    Returns:
      the snipped 2-points stats

    >>> two_point_stats(
    ...     da.from_array(np.arange(10).reshape(2, 5), chunks=(2, 5)),
    ...     da.from_array(np.arange(10).reshape(2, 5), chunks=(2, 5)),
    ... ).shape
    (2, 5)

    """
    cutoff_ = int((np.min(arr1.shape[1:]) - 1) / 2)
    if cutoff is None:
        cutoff = cutoff_
    cutoff = min(cutoff, cutoff_)

    nonperiodic_padder = sequence(
        dapad(
            pad_width=[(0, 0)] + [(cutoff, cutoff)] * (arr1.ndim - 1),
            mode="constant",
            constant_values=0,
        ),
        lambda x: da.rechunk(x, (x.chunks[0],) + x.shape[1:]),
    )

    padder = identity if periodic_boundary else nonperiodic_padder

    nonperiodic_normalize = lambda x: x / auto_correlation(padder(np.ones_like(arr1)))

    normalize = identity if periodic_boundary else nonperiodic_normalize

    return sequence(
        map_(padder),
        list,
        star(cross_correlation),
        normalize,
        center_slice(cutoff=cutoff),
    )([arr1, arr2])


@make_da
def correlations_multiple(data, correlations, periodic_boundary=True, cutoff=None):
    """Calculate 2-point stats for a multiple auto/cross correlation

    Args:
      data: the discretized data (n_samples,n_x,n_y,n_correlation)
      correlation_pair: the correlation pairs
      periodic_boundary: whether to assume a periodic boudnary (default is true)
      cutoff: the subarray of the 2 point stats to keep

    Returns:
      the 2-points stats array

    >>> data = np.arange(18).reshape(1, 3, 3, 2)
    >>> out = correlations_multiple(data, [[0, 1], [1, 1]])
    >>> out
    dask.array<stack, shape=(1, 3, 3, 2), dtype=float64, chunksize=(1, 3, 3, 1)>
    >>> answer = np.array([[[58, 62, 58], [94, 98, 94], [58, 62, 58]]]) + 1. / 3.
    >>> assert(out.compute()[...,0], answer)
    """

    return pipe(
        range(data.shape[-1]),
        map_(lambda x: (0, x)),
        lambda x: correlations if correlations else x,
        map_(
            lambda x: two_point_stats(
                data[..., x[0]],
                data[..., x[1]],
                periodic_boundary=periodic_boundary,
                cutoff=cutoff,
            )
        ),
        list,
        lambda x: da.stack(x, axis=-1),
        lambda x: da.rechunk(x, x.chunks[:-1] + (-1,)),
    )


class TwoPointCorrelation(BaseEstimator, TransformerMixin):
    """Calculate the 2-point stats for two arrays
    """

    def __init__(self, periodic_boundary=True, cutoff=None, correlations=None):
        """Instantiate a TwoPointCorrelation

        Args:
          periodic_boundary: whether the boundary conditions are periodic
          cutoff: cutoff radius of interest for the 2PtStatistics field
          correlations1: an index
          correlations2: an index

        """

        self.correlations = correlations
        self.periodic_boundary = periodic_boundary
        self.cutoff = cutoff

    def transform(self, data):
        """Transform the data

        Args:
          data: the data to be transformed

        Returns:
          the 2-point stats array
        """
        return correlations_multiple(
            data,
            self.correlations,
            periodic_boundary=self.periodic_boundary,
            cutoff=self.cutoff,
        )

    def fit(self, *_):
        """Only necessary to make pipelines work
        """
        return self


class FlattenTransformer(BaseEstimator, TransformerMixin):
    """Reshape data ready for a PCA.

    Two point correlation data need to be flatten before performing
    PCA. This class flattens the two point correlation data for use in
    a Sklearn pipeline.

    >>> data = np.arange(50).reshape((2, 5, 5))
    >>> FlattenTransformer().transform(data).shape
    (2, 25)

    """

    @staticmethod
    def transform(x_data):
        """Transform the X data

        Args:
          x_data: the data to be transformed

        """
        return flatten(x_data)

    def fit(self, *_):
        """Only necessary to make pipelines work
        """
        return self



#!/usr/bin/env python

"""PyMKS - the materials knowledge system in Python

See the documenation for details at https://pymks.org
"""

import warnings
import os
import subprocess
from setuptools import setup, find_packages


def make_version(package_name):
    """Generates a version number using `git describe`.

    Returns:
      version number of the form "3.1.1.dev127+g413ed61".
    """

    def _minimal_ext_cmd(cmd):
        """Run a command in a subprocess.

        Args:
          cmd: list of the command

        Returns:
          output from the command
        """
        # construct minimal environment
        env = {}
        for k in ["SYSTEMROOT", "PATH"]:
            value = os.environ.get(k)
            if value is not None:
                env[k] = value
        # LANGUAGE is used on win32
        env["LANGUAGE"] = "C"
        env["LANG"] = "C"
        env["LC_ALL"] = "C"
        out = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env).communicate()[0]
        return out

    version = "unknown"

    if os.path.exists(".git"):
        try:
            out = _minimal_ext_cmd(["git", "describe", "--tags", "--match", "v*"])
            # ticket:475 - fix for bytecode received in Py3k
            # http://jeetworks.org/node/67
            outdecode = out.decode("utf-8")
            # convert git long-form version string, e.g.,
            # "version-3_1_1-127-g413ed61", into PEP 440 version,
            # e.g., "3.1.1.dev127+g413ed61"
            version = outdecode.strip().split("-")
            if len(version) > 1:
                version, dev, sha = version
                version = "%s.dev%s+%s" % (version[1:], dev, sha)
            else:
                version = version[0][1:]
        except OSError:
            warnings.warn("Could not run ``git describe``")
    elif os.path.exists("pymks.egg-info"):
        # pylint: disable=import-outside-toplevel
        from pkg_resources import get_distribution, DistributionNotFound

        try:
            version = get_distribution(
                package_name
            ).version  # pylint: disable=no-member
        except DistributionNotFound:  # pragma: no cover
            version = "unknown, try running `python setup.py egg_info`"

    return version


PACKAGE_NAME = "pymks"


setup(
    name=PACKAGE_NAME,
    version=make_version(PACKAGE_NAME),
    description="Materials Knowledge Systems in Python (PyMKS)",
    author="Daniel Wheeler",
    author_email="daniel.wheeler2@gmail.com",
    url="http://pymks.org",
    packages=find_packages(),
    package_data={"": ["tests/*.py"]},
    install_requires=[
        "scipy",
        "dask",
        "scikit-learn",
        "toolz",
        "coverage",
        "matplotlib",
        "numpy",
        "pytest",
        "pytest-cov",
        "notebook",
        "dask-ml",
        "nbval",
        "graphviz",
        "Deprecated",
    ],
    data_files=["setup.cfg"],
)
